from ..imports import *


class MultiRainbow:
    """
    MultiRainbow objects collect multiple Rainbow objects together,
    providing a quick interface to apply the same action or visualization
    to all of them at once. It's meant to be a tool to facilitate
    quick comparisons between different pipeline analyses of the
    same dataset.
    """

    def __init__(self, list_of_rainbows, names=None):
        """
        Initialize from a list of Rainbows.

        Parameters
        ----------
        list_of_rainbows : list
            A list containing 1 or more Rainbow objects.
        names : list
            A list of names with which to label the Rainbows.
        """

        # store the Rainbow objects
        self.rainbows = list_of_rainbows

        # the list of names associated with those objects
        self.names = names

        # make sure the names and rainbows match up
        if self.names is not None:
            assert len(self.names) == len(self.rainbows)

    def __repr__(self):
        """
        How should this object be represented?
        """
        return f"<MultiRainbow({self.rainbows})>"

    @property
    def nrainbows(self):
        """
        What's the number of rainbows in here?
        """
        return len(self.rainbows)

    def _setup_panels(self, rows=1, figsize=None):
        """
        Set up a grid of panels to plot into.

        Parameters
        ----------
        rows : int
            The number of rows into which to arrange the
            plot panels. (This will probably only be needed
            for working with more than about 3 Rainbows.)
        figsize : tuple
            The size (width, height) for the figure to create.
            If left blank, this will be estimated from the
            default matplotlib figsize and the number of
            figures being produced.
        """

        # estimate a figure size from current matplotlib defaults
        if figsize == None:
            default_figsize = plt.matplotlib.rcParams["figure.figsize"]
            figsize = [self.nrainbows * default_figsize[0], default_figsize[1]]

        # create the figure and grid of axes as subplots
        self.figure, self.axes = plt.subplots(
            rows,
            int(np.ceil(self.nrainbows / rows)),
            sharex=True,
            sharey=True,
            figsize=figsize,
        )

        # give titles to the plot panels
        if self.names is not None:
            for name, ax in zip(self.names, self.axes):
                ax.set_title(name)

    def _guess_good_uniform_wavelength_grid(
        self, fraction_to_supersample=0.5, wscale=None, plot=False
    ):
        """
        Guess a good shared wavelength grid that would work
        reasonably well for all the Rainbows in this set.

        Parameters
        ----------
        fraction_to_supersample : float
            When setting the new wavelength bins, what fraction of bins
            in the original Rainbows will be super-sampled? 0.0 will create
            new bins that are just larger than the largest original bin,
            1.0 will create new bins that are just smaller than the smallest
            original bin, and 0.5 is the default.

        plot : bool
            Should we make a plot of the d[wavelength]/d[bin], for diagnostics?
        """

        # pick a unit to make sure everything's consistent
        w_unit = u.micron

        # create a list of interpolating functions for the datasets' dw
        dw_interpolators = [
            interp1d(
                r.wavelength.to(w_unit).value,
                np.gradient(r.wavelength.to(w_unit).value),
                fill_value=np.inf,
                bounds_error=False,
            )
            for r in self.rainbows
        ]

        def smallest_dw(w):
            """
            Helper function to determine the smallest dw
            in the datasets at any particular wavelength.

            Parameters
            ----------
            w : np.array
                Wavelength (with no units).
            """

            dw = np.inf * np.ones_like(w)
            for interpolator in dw_interpolators:
                dw = np.minimum(dw, interpolator(w))
            return dw

        # compile w and dw, weighted by how many appear in each dataset
        w_represented = np.sort(
            np.hstack([r.wavelength.to(w_unit).value for r in self.rainbows])
        )
        dw_represented = smallest_dw(w_represented)

        # fit a line to log(dw) as a function of log(w)
        slope, intercept = np.polyfit(np.log(w_represented), np.log(dw_represented), 1)
        dw_model = np.exp(np.polyval([slope, intercept], np.log(w_represented)))

        # decide if it should be linear or logarithmic
        if wscale is None:
            if slope < 0.5:
                wscale = "linear"
            else:
                wscale = "log"

        if wscale == "linear":
            dw = np.percentile(dw_represented, 100 * (1 - fraction_to_supersample))
            new_wavelengths = np.arange(
                np.nanmin(w_represented), np.nanmax(w_represented), dw
            )
        elif wscale == "log":
            R = np.percentile(
                w_represented / dw_represented, 100 * fraction_to_supersample
            )
            new_wavelengths = np.exp(
                np.arange(
                    np.log(np.nanmin(w_represented)),
                    np.log(np.nanmax(w_represented)),
                    1 / R,
                )
            )
        else:
            raise RuntimeError('🌈 Currently only "log" or "linear" are supported.')

        if plot:

            # plot the individual datasets
            for r in self.rainbows:
                plt.plot(
                    r.wavelength, np.gradient(r.wavelength), marker="o", alpha=0.25
                )

            # plot the minimum dw at each wavelength
            plt.plot(
                w_represented, dw_represented, marker="o", color="black", alpha=0.25
            )

            # convert to log scale, so R=constant is a slope=1 line
            plt.yscale("log")
            plt.xscale("log")
            plt.axis("scaled")

            # plot the line fit result
            plt.plot(
                w_represented,
                dw_model,
                linestyle="--",
                color="black",
                linewidth=2,
                alpha=0.25,
            )

            # plot the new logarithmic or linear wavelength grid
            plt.plot(
                new_wavelengths,
                np.gradient(new_wavelengths),
                color="black",
                linewidth=20,
                alpha=0.25,
            )
            plt.title(f"{slope:.3} -> {wscale}!")
            plt.ylabel("d[wavelength]/d[bin]")

        return new_wavelengths * w_unit

    def _check_if_wavelengths_are_aligned(self):
        """
        Check whether the wavelength grid is already aligned.

        Returns
        -------
        aligned : bool
            Are the wavelengths the same across all?
        """
        # check if they're already aligned
        wavelengths_are_same_size = np.all(
            [np.all(r.nwave == self.rainbows[0].nwave) for r in self.rainbows]
        )

        if wavelengths_are_same_size:
            wavelengths_are_already_aligned = np.all(
                [
                    np.all(r.wavelength == self.rainbows[0].wavelength)
                    for r in self.rainbows
                ]
            )
        else:
            wavelengths_are_already_aligned = False

        return wavelengths_are_already_aligned

    def _check_if_times_are_aligned(self):
        """
        Check whether the time grid is already aligned.

        Returns
        -------
        aligned : bool
            Are the times the same across all?
        """
        # check if they're already aligned
        times_are_same_size = np.all(
            [np.all(r.ntime == self.rainbows[0].ntime) for r in self.rainbows]
        )

        if times_are_same_size:
            times_are_already_aligned = np.all(
                [np.all(r.time == self.rainbows[0].wavelength) for r in self.rainbows]
            )
        else:
            times_are_already_aligned = False

        return times_are_already_aligned

    @property
    def wavelength(self):
        if self._check_if_wavelengths_are_aligned():
            return self.rainbows[0].wavelength
        else:
            raise RuntimeError(f"🌈 {self} has more than one wavelength array.")

    @property
    def time(self):
        if self._check_if_times_are_aligned():
            return self.rainbows[0].time
        else:
            raise RuntimeError(f"🌈 {self} has more than one time array.")

    def align_wavelengths(self, **kw):
        """
        Make a new MultiRainbow where wavelengths have
        been aligned across all the Rainbows, making
        a guesst for whether
        """

        if self._check_if_wavelengths_are_aligned():
            return self
        else:
            w = self._guess_good_uniform_wavelength_grid(**kw)
            return self.bin(wavelength=w)

    def __getitem__(self, key):
        """
        Trim a rainbow by indexing, slicing, or masking.
        Two indices must be provided (`[:,:]`).

        Examples
        --------
        ```
        r[:,:]
        r[10:20, :]
        r[np.arange(10,20), :]
        r[r.wavelength > 1*u.micron, :]
        r[:, np.abs(r.time) < 1*u.hour]
        r[r.wavelength > 1*u.micron, np.abs(r.time) < 1*u.hour]
        ```

        Parameters
        ----------
        key : tuple
            The (wavelength, time) slices, indices, or masks.
        """
        new_rainbows = [r.__getitem__(key) for r in self.rainbows]
        return MultiRainbow(new_rainbows, names=self.names)

    def normalize(self, **kwargs):
        """
        Normalize by dividing through by the median spectrum and/or lightcurve.

        Parameters
        ----------
        wavelength : bool
            Should we divide by the median spectrum?

        time : bool
            Should we divide by the median light curve?

        Returns
        -------
        normalized : MultiRainbow
            The normalized MultiRainbow.
        """
        new_rainbows = [r.normalize(**kwargs) for r in self.rainbows]
        return MultiRainbow(new_rainbows, names=self.names)

    def bin(self, **kwargs):
        """
        Bin the rainbow in wavelength and/or time.

        The time-setting order of precendence is
        [`time`, `dt`], meaning that if `time` is set,
        any values given for `dt` will be ignored.
        The wavelength-setting order of precendence is
        [`wavelength`, `dw`, `R`], meaning that if `wavelength`
        is set any values of `dw` or `R` will be ignored, and
        if `dw` is set any value of `R` will be ignored.

        Parameters
        ----------
        dt : astropy.units.Quantity
            The d(time) bin size for creating a grid
            that is uniform in linear space.
        time : array of astropy.units.Quantity
            An array of times, if you just want to give
            it an entirely custom array.
        R : float
            The spectral resolution for creating a grid
            that is uniform in logarithmic space.
        dw : astropy.units.Quantity
            The d(wavelength) bin size for creating a grid
            that is uniform in linear space.
        wavelength : array of astropy.units.Quantity
            An array of wavelengths, if you just want to give
            it an entirely custom array.

        Returns
        -------
        binned : MultiRainbow
            The binned MultiRainbow.
        """
        new_rainbows = [r.bin(**kwargs) for r in self.rainbows]
        return MultiRainbow(new_rainbows, names=self.names)

    def plot(self, **kwargs):
        """
        Plot flux as sequence of offset light curves.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            The axes into which to make this plot.
        spacing : None, float
            The spacing between light curves.
            (Might still change how this works.)
            None uses half the standard dev of entire flux data.
        w_unit : str, astropy.unit.Unit
            The unit for plotting wavelengths.
        t_unit : str, astropy.unit.Unit
            The unit for plotting times.
        cmap : str, matplotlib.colors.Colormap
            The color map to use for expressing wavelength.
        vmin : astropy.units.Quantity
            The minimum value to use for the wavelength colormap.
        vmax : astropy.units.Quantity
            The maximum value to use for the wavelength colormap.
        plowkw : dict
            A dictionary of keywords passed to `plt.plot`
        textkw : dict
            A dictionary of keywords passed to `plt.text`
        """

        # set up a grid of panels
        self._setup_panels()

        # make all the individual plots
        for r, a in zip(self.rainbows, self.axes):
            r.plot(ax=a, **kwargs)

    def imshow(self, vmin=None, vmax=None, **kwargs):
        """
        imshow flux as a function of time (x = time, y = wavelength, color = flux).

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            The axes into which to make this plot.
        quantity : str
            The fluxlike quantity to imshow.
            (Must be a key of `rainbow.fluxlike`).
        w_unit : str, astropy.unit.Unit
            The unit for plotting wavelengths.
        t_unit : str, astropy.unit.Unit
            The unit for plotting times.
        colorbar : bool
            Should we include a colorbar?
        aspect : str
            What aspect ratio should be used for the imshow?
        """

        # set up a grid of panels
        self._setup_panels()

        # figure out a good shared color limits
        vmin = vmin or np.nanmin(
            [np.nanmin(u.Quantity(r.flux).value) for r in self.rainbows]
        )
        vmax = vmax or np.nanmax(
            [np.nanmax(u.Quantity(r.flux).value) for r in self.rainbows]
        )

        # make all the individual imshows
        for r, a in zip(self.rainbows, self.axes):
            r.imshow(ax=a, vmin=vmin, vmax=vmax, **kwargs)

    def animate_lightcurves(
        self,
        filename="multi-animated-lightcurves.gif",
        fps=None,
        dpi=None,
        **kwargs,
    ):
        """
        Create an animation to show how the lightcurve changes
        as we flip through every wavelength.

        Parameters
        ----------
        filename : str
            Name of file you'd like to save results in.
            Currently supports only .gif files.
        fps : float
            frames/second of animation
        ax : matplotlib.axes.Axes
            The axes into which this animated plot should go.
        xlim : tuple
            Custom xlimits for the plot
        ylim : tuple
            Custom ylimits for the plot
        cmap : str, matplotlib.colors.Colormap
            The color map to use for expressing wavelength
        vmin : astropy.units.Quantity
            The minimum value to use for the wavelength colormap
        vmax : astropy.units.Quantity
            The maximum value to use for the wavelength colormap
        scatterkw : dict
            A dictionary of keywords to be passed to `plt.scatter`
        textkw : dict
            A dictionary of keywords to be passed to `plt.text`
        """

        assert np.all([r.nwave == self.rainbows[0].nwave for r in self.rainbows])

        # set up a grid of panels
        self._setup_panels()

        # setup all the individual plots
        for r, a in zip(self.rainbows, self.axes):
            r._setup_animate_lightcurves(ax=a, **kwargs)

        # the figure should be the same for all panels
        figure = r._animate_lightcurves_components["fig"]

        def update(frame):
            """
            This function will be called to update each frame
            of the animation.

            Parameters
            ----------
            frame : int
                An integer that will advance with each frame.
            """
            artists = []
            for r in self.rainbows:
                r._animate_lightcurves_components["update"](frame)
                for k in ["text", "scatter"]:
                    artists.append(r._animate_lightcurves_components[k])
            return artists

        # make and save the animation
        animator = ani.FuncAnimation(
            figure,
            update,
            frames=np.arange(0, r.nwave),
            blit=True,
        )
        animator.save(
            filename, fps=fps, dpi=dpi, savefig_kwargs=dict(facecolor="white")
        )
        plt.close()

    def animate_spectra(
        self, filename="multi-animated-spectra.gif", fps=None, dpi=None, **kwargs
    ):
        """
        Create an animation to show how the spectrum changes
        as we flip through every timepoint.

        Parameters
        ----------
        filename : str
            Name of file you'd like to save results in.
            Currently supports only .gif files.
        fps : float
            frames/second of animation
        ax : matplotlib.axes.Axes
            The axes into which this animated plot should go.
        xlim : tuple
            Custom xlimits for the plot
        ylim : tuple
            Custom ylimits for the plot
        cmap : str, matplotlib.colors.Colormap
            The color map to use for expressing wavelength
        vmin : astropy.units.Quantity
            The minimum value to use for the wavelength colormap
        vmax : astropy.units.Quantity
            The maximum value to use for the wavelength colormap
        scatterkw : dict
            A dictionary of keywords to be passed to `plt.scatter`
        textkw : dict
            A dictionary of keywords to be passed to `plt.text`
        """

        assert np.all([r.ntime == self.rainbows[0].ntime for r in self.rainbows])

        # set up a grid of panels
        self._setup_panels()

        # setup all the individual plots
        for r, a in zip(self.rainbows, self.axes):
            r._setup_animate_spectra(ax=a, **kwargs)

        # the figure should be the same for all panels
        figure = r._animate_spectra_components["fig"]

        def update(frame):
            """
            This function will be called to update each frame
            of the animation.

            Parameters
            ----------
            frame : int
                An integer that will advance with each frame.
            """
            artists = []
            for r in self.rainbows:
                r._animate_spectra_components["update"](frame)
                for k in ["text", "scatter"]:
                    artists.append(r._animate_spectra_components[k])
            return artists

        # make and save the animation
        animator = ani.FuncAnimation(
            figure,
            update,
            frames=np.arange(0, r.ntime),
            blit=True,
        )
        animator.save(
            filename, fps=fps, dpi=dpi, savefig_kwargs=dict(facecolor="white")
        )
        plt.close()



# import the general list of packages
from ...imports import *

# define list of the only things that will show up in imports
__all__ = ["from_radica"]


def from_radica(rainbow, filepath, order=1):
    """
    Populate a Rainbow from a file in the radica format.

    Parameters
    ----------

    rainbow : Rainbow
        The object to be populated.
    filepath : str
        The path to the file to load.
    order : int
        The spectral order to be loaded.
    """

    # open the FITS file
    hdu_list = fits.open(filepath)

    # load the header into metadata
    rainbow.metadata = dict(hdu_list["primary"].header)

    # populate a 1D array of wavelengths (with astropy units of length)
    wavelength_2d = hdu_list[f"wo{order}"].data
    rainbow.wavelike["wavelength"] = np.median(wavelength_2d, 0) * u.micron

    # populate a 1D array of times (with astropy units of time)
    # populate a 1D array of times (with astropy units of time)
    times = np.arange(wavelength_2d.shape[0]) * u.minute
    warnings.warn("\nThe times are totally made up!")
    rainbow.timelike["time"] = times

    # populate a 2D (row = wavelength, col = array of fluxes
    rainbow.fluxlike["flux"] = hdu_list[f"fo{order}"].data.T
    rainbow.fluxlike["uncertainty"] = hdu_list[f"eo{order}"].data.T

    message = f"""
    Loading NIRISS order '{order}'. If you want the other order,
    trying `r = Rainbow(..., format='radica', order=2)`
    """
    warnings.warn(message)



from ...imports import *


def get_spectral_resolution(self, pixels_per_resolution_element=1):
    """
    Parameters
    ----------
    pixels_per_resolution_element : float
        How many pixels do we consider as a resolution element?

    Returns
    -------
    R : np.array (wavelike)
        The spectral resolution, at each wavelength.
    """

    # calculate spectral resolution, for this pixels/element
    w = self.wavelength
    dw = np.gradient(self.wavelength)
    R = np.abs(w / dw / pixels_per_resolution_element)

    return R


def plot_spectral_resolution(
    self,
    pixels_per_resolution_element=1,
    ax=None,
    w_unit="micron",
    cmap=None,
    vmin=None,
    vmax=None,
    plotkw={},
    **kw,
):
    """
    Plot the spectral resolution as a function of wavelength.

    Parameters
    ----------
    pixels_per_resolution_element : float
        How many pixels do we consider as a resolution element?
    ax : matplotlib.axes.Axes
        The axes into which to make this plot.
    w_unit : str, astropy.unit.Unit
        The unit for plotting wavelengths.
    cmap : str, matplotlib.colors.Colormap
        The color map to use for expressing wavelength.
    vmin : astropy.units.Quantity
        The minimum value to use for the wavelength colormap.
    vmax : astropy.units.Quantity
        The maximum value to use for the wavelength colormap.
    plotkw : dict
        A dictionary of keywords passed to `plt.plot`
    """
    R = self.get_spectral_resolution(pixels_per_resolution_element)
    w_unit = u.Unit(w_unit)

    with quantity_support():

        # make sure that the wavelength-based colormap is defined
        self._make_sure_cmap_is_defined(cmap=cmap, vmin=vmin, vmax=vmax)

        # make sure ax is set up
        if ax is None:
            ax = plt.subplot()
        plt.sca(ax)

        # define default visuals, but let user override with scatterkw
        # this_scatterkw = dict(c=self.get_wavelength_color(self.wavelength))
        # this_scatterkw.update(**scatterkw)

        plt.plot(self.wavelength.to(w_unit), R, **plotkw)
        plt.xlabel(f'Wavelength ({w_unit.to_string("latex_inline")})')
        plt.ylabel(f"$R=\lambda/d\lambda$ ({pixels_per_resolution_element} pixel)")
        return ax

        ax.set_xlabel(f"Time ({self.time.unit.to_string('latex_inline')})")


def get_typical_uncertainty(self, function=np.nanmedian):
    """
    Get the typical per-wavelength uncertainty.

    Parameters
    ----------
    function : function
        What function should be used to choose the "typical"
        value for each wavelength? Good options are probably
        things like `np.nanmedian`, `np.median`, `np.nanmean`
        `np.mean`

    Returns
    -------
    uncertainty_per_wavelength : np.array (wavelike)
        The uncertainty associated with each wavelength.
    """
    uncertainty_per_wavelength = function(self.uncertainty, axis=self.timeaxis)
    return uncertainty_per_wavelength


def get_spectrum(self):
    """
    Return a spectrum of the star, averaged over all times.

    Parameters
    ----------


    Returns
    -------
    s : np.array (wavelike)
    """

    # FIXME, think about error weighting/masking
    return np.nanmedian(self.flux, axis=self.timeaxis)



from ..imports import *


def __add__(self, object):
    """
    Add the flux of a rainbow and an input array (or another rainbow)
    and output in a new rainbow object.
    Currently only supports flux addition.

    Parameters
    ----------
    object : Array or float.
        Multiple options:
        1) float
        2) 1D array with same length as wavelength axis
        3) 1D array with same length as time axis
        4) 2D array with same shape as rainbow flux
        5) Rainbow object with same dimensions as self.

    Returns
    ----------
    rainbow : Rainbow() with the same parameters as self but with added
    flux.
    """

    # Create new Rainbow() to store results in.
    result = self._create_copy()

    try:  # Object is rainbow
        if np.array_equal(
            self.wavelike["wavelength"], object.wavelike["wavelength"]
        ) and np.array_equal(self.timelike["time"], object.timelike["time"]):
            result.fluxlike["flux"] += object.fluxlike["flux"]
        else:
            print("Objects do not share wavelength/time axes")
            return

    except AttributeError:  # Object is not rainbow

        if type(object) == float or type(object) == int:  # Float or integer.
            result.fluxlike["flux"] += object

        elif self.shape == object.shape:  # fluxlike
            result.fluxlike["flux"] += object

        elif len(object) == len(
            self.wavelike["wavelength"]
        ):  # Add along wavelength axis

            result.fluxlike["flux"] += np.transpose([object] * self.shape[1])
            if self.nwave == self.ntime:
                raise RuntimeError(
                    f"{self} has same number of wavelengths and times; we can't tell which is which."
                )

        elif len(object) == len(self.timelike["time"]):  # Add along time axis.
            result.fluxlike["flux"] += np.tile(object, (self.shape[0], 1))

        else:
            print("Invalid shape in addition: " + str(object.shape))
            return
    return result


def __sub__(self, object):
    """
    Subtract the flux of a rainbow from an input array (or another rainbow)
    and output in a new rainbow object.
    Currently only supports flux subtraction.

    Parameters
    ----------
    object : Array or float.
        Multiple options:
        1) float
        2) 1D array with same length as wavelength axis
        3) 1D array with same length as time axis
        4) 2D array with same shape as rainbow flux
        5) Rainbow object with same dimensions as self.

    Returns
    ----------
    rainbow : Rainbow() with the same parameters as self but with subtracted
    flux.
    """

    # Create new Rainbow() to store results in.
    result = self._create_copy()

    try:  # Object is rainbow.
        if np.array_equal(
            self.wavelike["wavelength"], object.wavelike["wavelength"]
        ) and np.array_equal(self.timelike["time"], object.timelike["time"]):
            result.fluxlike["flux"] -= object.fluxlike["flux"]
        else:
            print("Objects do not share wavelength/time axes")
            return

    except AttributeError:

        if type(object) == float or type(object) == int:  # Float or integer.
            result.fluxlike["flux"] -= object

        elif self.shape == object.shape:  # fluxlike
            result.fluxlike["flux"] -= object

        elif len(object) == len(
            self.wavelike["wavelength"]
        ):  # Add along wavelength axis
            result.fluxlike["flux"] -= np.transpose([object] * self.shape[1])
            if self.nwave == self.ntime:
                raise RuntimeError(
                    f"{self} has same number of wavelengths and times; we can't tell which is which."
                )

        elif len(object) == len(self.timelike["time"]):  # Add along time axis.
            result.fluxlike["flux"] -= np.tile(object, (self.shape[0], 1))

        else:
            print("Invalid shape in subtraction: " + str(object.shape))
            return
    return result


def __mul__(self, object):
    """
    Multiply the flux of a rainbow and an input array (or another rainbow)
    and output in a new rainbow object.
    Currently only supports flux multiplication.

    Parameters
    ----------
    object : Array or float.
        Multiple options:
        1) float
        2) 1D array with same length as wavelength axis
        3) 1D array with same length as time axis
        4) 2D array with same shape as rainbow flux
        5) Rainbow object with same dimensions as self.

    Returns
    ----------
    rainbow : Rainbow() with the same parameters as self but with multiplied
    flux.
    """

    # Create new Rainbow() to store results in.
    result = self._create_copy()

    try:  # Object is rainbow
        if np.array_equal(
            self.wavelike["wavelength"], object.wavelike["wavelength"]
        ) and np.array_equal(self.timelike["time"], object.timelike["time"]):
            result.fluxlike["flux"] *= object.fluxlike["flux"]
        else:
            print("Objects do not share wavelength/time axes")
            return

    except AttributeError:

        if type(object) == float or type(object) == int:  # Float or integer.
            result.fluxlike["flux"] *= object

        elif self.shape == object.shape:  # fluxlike
            result.fluxlike["flux"] *= object

        elif len(object) == len(
            self.wavelike["wavelength"]
        ):  # Add along wavelength axis
            result.fluxlike["flux"] *= np.transpose([object] * self.shape[1])
            if self.nwave == self.ntime:
                raise RuntimeError(
                    f"{self} has same number of wavelengths and times; we can't tell which is which."
                )

        elif len(object) == len(self.timelike["time"]):  # Add along time axis.
            result.fluxlike["flux"] *= np.tile(object, (self.shape[0], 1))

        else:
            print("Invalid shape in multiplication: " + str(object.shape))
            return
    return result


def __truediv__(self, object):
    """
    Divide the flux of a rainbow and an input array (or another rainbow)
    and output in a new rainbow object.
    Currently only supports flux division.

    Parameters
    ----------
    object : Array or float.
        Multiple options:
        1) float
        2) 1D array with same length as wavelength axis
        3) 1D array with same length as time axis
        4) 2D array with same shape as rainbow flux
        5) Rainbow object with same dimensions as self.

    Returns
    ----------
    rainbow : Rainbow() with the same parameters as self but with divided
    flux.
    """

    # Create new Rainbow() to store results in.
    result = self._create_copy()

    try:  # Object is another rainbow.
        if np.array_equal(
            self.wavelike["wavelength"], object.wavelike["wavelength"]
        ) and np.array_equal(self.timelike["time"], object.timelike["time"]):
            result.fluxlike["flux"] /= object.fluxlike["flux"]
        else:
            print("Objects do not share wavelength/time axes")
            return

    except AttributeError:

        if type(object) == float or type(object) == int:  # float
            result.fluxlike["flux"] /= object

        elif self.shape == object.shape:  # fluxlike array
            result.fluxlike["flux"] /= object

        elif len(object) == len(
            self.wavelike["wavelength"]
        ):  # Add along wavelength axis
            result.fluxlike["flux"] /= np.transpose([object] * self.shape[1])
            if self.nwave == self.ntime:
                raise RuntimeError(
                    f"{self} has same number of wavelengths and times; we can't tell which is which."
                )
        elif len(object) == len(self.timelike["time"]):  # Add along time axis.
            result.fluxlike["flux"] /= np.tile(object, (self.shape[0], 1))

        else:
            print("Invalid shape in division: " + str(object.shape))
            return
    return result


def __eq__(self, other):
    """
    Test whether (self) == (other).
    """
    # start by assuming the Rainbows are identical
    same = True

    # loop through the core dictionaries
    for d in self._core_dictionaries:

        # pull out each core dictionary from both
        d1, d2 = vars(self)[d], vars(other)[d]

        # loop through elements of each dictionary
        for k in d1:

            # test that all elements match for both
            same *= np.all(d1[k] == d2.get(k, None))

    return same



from ..rainbows import *
from .setup_tests import *


def test_bin_in_time():

    s = SimulatedRainbow(dt=5 * u.minute, signal_to_noise=100)
    b = s.bin_in_time(dt=1 * u.hour)
    assert b.ntime < s.ntime

    assert s.bin_in_time(dt=None, time=None) == s

    fi, ax = plt.subplots(2, 1, sharex=True)
    imshowkw = dict(vmin=0.98, vmax=1.02)
    s.imshow(ax=ax[0], **imshowkw)
    b.imshow(ax=ax[1], **imshowkw)
    plt.savefig(os.path.join(test_directory, "imshow-bin-time-demonstration.pdf"))


def test_bin_in_wavelength():

    s = SimulatedRainbow(dw=50 * u.nm)
    b = s.bin_in_wavelength(dw=500 * u.nm)
    assert b.nwave < s.nwave

    assert s.bin_in_wavelength(R=None, dw=None, wavelength=None) == s

    fi, ax = plt.subplots(2, 1, sharex=True)
    imshowkw = dict(vmin=0.98, vmax=1.02)
    s.imshow(ax=ax[0], **imshowkw)
    b.imshow(ax=ax[1], **imshowkw)
    plt.savefig(os.path.join(test_directory, "imshow-bin-wavelength-demonstration.pdf"))


def test_bin():

    s = SimulatedRainbow(dt=10 * u.minute, R=25)
    b = s.bin(dt=0.5 * u.hour, dw=0.5 * u.micron)
    assert b.nwave < s.nwave
    assert b.ntime < s.ntime

    assert s.bin(dw=None, dt=None, R=None, time=None, wavelength=None) == s

    fi, ax = plt.subplots(2, 1, sharex=True)
    imshowkw = dict(vmin=0.98, vmax=1.02)
    s.imshow(ax=ax[0], **imshowkw)
    plt.title("Unbinned")
    b.imshow(ax=ax[1], **imshowkw)
    plt.title("Binned")
    plt.tight_layout()
    plt.savefig(os.path.join(test_directory, "imshow-bin-demonstration.pdf"))


def test_bin_uncertainty_basic(original_resolution=100, binned_resolution=42):

    a = SimulatedRainbow(R=original_resolution, signal_to_noise=100)
    b = a.bin(R=binned_resolution)
    original_uncertainty = np.median(a.uncertainty)
    predicted_uncertainty = np.median(a.uncertainty) / np.sqrt(
        original_resolution / binned_resolution
    )
    actual_uncertainty = np.median(b.uncertainty)

    did_it_work = np.isclose(predicted_uncertainty, actual_uncertainty, rtol=0.001)

    print(
        f"""
    At the original resolution of R={original_resolution},
    the median uncertainty was {original_uncertainty}.
    When binned to R={binned_resolution}, the predicted
    resolution should be {predicted_uncertainty}.
    It is actually {actual_uncertainty},
    so it is {did_it_work} that the binning worked!
    """
    )

    assert did_it_work


def test_bin_bad_data(visualize=False):
    s = SimulatedRainbow(signal_to_noise=100, R=20, dt=5 * u.minute)
    N = 100
    i, j = np.random.randint(0, s.nwave, N), np.random.randint(0, s.ntime, N)
    s.fluxlike["flux"][i, j] = np.nan

    i, j = np.random.randint(0, s.nwave, N), np.random.randint(0, s.ntime, N)
    s.fluxlike["uncertainty"][i, j] = 0

    b = s.bin(R=7, dt=15 * u.minute)

    if visualize:
        fi, ax = plt.subplots(2, 2, figsize=(8, 6), dpi=300, sharex=True, sharey=True)
        s.imshow(ax=ax[0, 0], quantity="uncertainty", vmax=0.01)
        s.imshow(ax=ax[0, 1], vmin=0.97, vmax=1.03)

        b.imshow(ax=ax[1, 0], quantity="uncertainty", vmax=0.01)
        b.imshow(ax=ax[1, 1], vmin=0.97, vmax=1.03)

    assert np.any(np.isfinite(b.flux))
    return s


def test_normalize(plot=False):
    N = 37
    w = np.logspace(0, 1, N) * u.micron
    f = np.cos(2 * np.pi * w.value / 3) + 1

    snr = 100
    a = SimulatedRainbow(signal_to_noise=snr, wavelength=w)
    b = SimulatedRainbow(signal_to_noise=snr, wavelength=w, star_flux=f)
    c = SimulatedRainbow(
        signal_to_noise=snr, wavelength=w, star_flux=f
    ).inject_transit()

    for x in [a, b, c]:
        nw = x.normalize(axis="w")
        nt = x.normalize(axis="t")
        nwt = x.normalize(axis="w").normalize(axis="t")
        ntw = x.normalize(axis="t").normalize(axis="w")

        for r in [nw, nt, nwt, ntw]:
            r.fluxlike["relative-uncertainty"] = r.uncertainty / r.flux
            assert np.all(np.isclose(r.uncertainty / r.flux, 1 / snr, rtol=0.1))
            if plot:
                r.imshow_fluxlike_quantities(maxcol=4)



from ..rainbows import *
from .setup_tests import *


def test_multi():
    a = SimulatedRainbow(signal_to_noise=400).inject_transit()
    b = SimulatedRainbow(signal_to_noise=800).inject_transit()

    m = MultiRainbow([a, b])
    m.bin(R=5).plot(spacing=0.02)
    plt.savefig(os.path.join(test_directory, "multi-plot-demonstration.png"))

    m.imshow()
    plt.savefig(os.path.join(test_directory, "multi-imshow-demonstration.png"))

    m.animate_lightcurves(
        filename=os.path.join(
            test_directory, "multi-animate-lightcurves-demonstration.gif"
        )
    )
    m.animate_spectra(
        filename=os.path.join(test_directory, "multi-animate-spectra-demonstration.gif")
    )

    m.normalize()
    m.align_wavelengths().wavelength
    m[:, :]



from .test_io import *
from .test_rainbow_actions import *
from .test_rainbow_basics import *
from .test_rainbow_operations import *
from .test_rainbow_simulations import *
from .test_rainbow_visualizations import *



from ...imports import *


def normalize(self, axis="wavelength", percentile=50):
    """
    Normalize by dividing through by the median spectrum and/or lightcurve.

    Parameters
    ----------
    axis : str
        The axis that should be normalized out.
        `w` or `wave` or `wavelength` will divide out the typical spectrum.
        `t` or `time` will divide out the typical light curve

    percentile : float
        A number between 0 and 100, specifying the percentile
        of the data along an axis to use as the reference.
        The default of `percentile=50` corresponds to the median.
        If you want to normalize to out-of-transit, maybe you
        want a higher percentile. If you want to normalize to
        the baseline below a flare, maybe you want a lower
        percentage.

    Returns
    -------
    normalized : MultiRainbow
        The normalized MultiRainbow.
    """

    # TODO, think about more careful treatment of uncertainties + good/bad data
    new = self._create_copy()

    # (ignore nan warnings)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        if axis.lower()[0] == "w":
            normalization = np.nanpercentile(new.flux, percentile, axis=self.timeaxis)
            new.fluxlike["flux"] = new.flux / normalization[:, np.newaxis]
            try:
                new.fluxlike["uncertainty"] = (
                    self.uncertainty / normalization[:, np.newaxis]
                )
            except ValueError:
                pass
        elif axis.lower()[0] == "t":
            normalization = np.nanpercentile(self.flux, percentile, axis=self.waveaxis)
            new.fluxlike["flux"] = new.flux / normalization[np.newaxis, :]
            try:
                new.fluxlike["uncertainty"] = (
                    self.uncertainty / normalization[np.newaxis, :]
                )
            except ValueError:
                pass

    return new



"""
Tools for resampling array from one grid
of independent variables to another.
"""

from .imports import *

__all__ = ["bintoR", "bintogrid"]


def binsizes(x):
    """
    If x is an array of bin centers, calculate the bin sizes.
    (assumes outermost bins are same size as their neighbors)

    Parameters
    ----------
    x : np.array
        The array of bin centers.

    Returns
    ----------
    s : np.array
        The array of bin sizes (total size, from left to right).
    """

    binsize = np.zeros_like(x)
    binsize[0:-1] = x[1:] - x[0:-1]
    binsize[-1] = binsize[-2]
    return binsize


def binedges(x):
    """
    If x is an array of bin centers, calculate the bin edges.
    (assumes outermost bins are same size as their neighbors)

    Parameters
    ----------
    x : np.array
        The array of bin centers.

    Returns
    ----------
    l : np.array
        The left edges of the bins.
    r : np.array
        The right edges of the bins.
    """

    # what are bin edges (making a guess for those on the ends)
    xbinsize = binsizes(x)
    xleft = x - xbinsize / 2.0
    xright = x + xbinsize / 2.0

    return xleft, xright


def plotboxy(x, y, **kwargs):
    """
    Plot with boxes, to show the left and right edges of a box.
    This is useful, or example, to plot flux associated with
    pixels, in case you are trying to do a sub-pixel resample
    or interpolation or shift.

    Parameters
    ----------
    x : np.array
        The original independent variable.
    y : np.array
        The original dependent variable (same size as x).
    **kwargs : dict
        All additional keywords will be passed to plt.plot
    """

    # what are bin edges (making a guess for those on the ends)
    xleft, xright = binedges(x)

    # create a array doubling up the y values and interleaving the edges
    plot_x = np.vstack((xleft, xright)).reshape((-1,), order="F")
    plot_y = np.vstack((y, y)).reshape((-1,), order="F")

    # plot those constructed arrays
    plt.plot(plot_x, plot_y, **kwargs)


def fluxconservingresample(
    xin_unsorted,
    yin_unsorted,
    xout,
    replace_nans=0.0,
    visualize=False,
    pause=False,
):
    """
    Starting from some initial x and y, resample onto a
    different grid (either higher or lower resolution),
    while conserving total flux.

    When including the entire range of xin,
    sum(yout) == sum(yin) should be true.

    When including only part of the range of xin,
    the integral between any two points should be conserved.

    Parameters
    ----------

    xin_unsorted : np.array
        The original independent variable.
        It doesn't necessarily have to be sorted.
    yin_unsorted : np.array
        The original dependent variable (same size as x).
        It doesn't necessarily have to be sorted.
    xout : np.array
        The new grid of independent variables onto which
        you want to resample the y values.
    replace_nans : float, str
        Replace nan values with this.
        `replace_nans = 0`
            will add no flux where nans are
        `replace_nans = nan`
            will ensure you get nans returned everywhere
            if you try to resample over any nan
        `replace_nans = 'interpolate'`
            will try to replace nans by linearly interpolating
            from nearby values (not yet implemented)
    visualize : bool
        Should we make a plot showing it worked?
    pause : bool
        Should we pause after making that plot?
    """

    # sort to make sure x is strictly increasing
    s = np.argsort(xin_unsorted)
    xin = xin_unsorted[s]
    yin = yin_unsorted[s]

    # set up the bins, to calculate cumulative distribution of y
    xinleft, xinright = binedges(xin)

    # the first element should be the left edge of the first pixel
    # last element will be right edge of last pixel
    xinforcdf = np.hstack([xinleft, xinright[-1]])

    # to the left of the first pixel, assume flux is zero
    yinforcdf = np.hstack([0, yin])

    # correct for any non-finite values
    bad = np.isnan(yinforcdf)
    if replace_nans == "interpolate":
        raise NotImplementedError(
            "The `replace_nans='interpolate'`` option doens't exist yet!"
        )
    yinforcdf[bad] = replace_nans

    # calculate the CDF of the flux (at pixel edge locations)
    cdfin = np.cumsum(yinforcdf)

    # create an interpolator for that CDF
    cdfinterpolator = interp1d(
        xinforcdf,
        cdfin,
        kind="linear",
        bounds_error=False,
        fill_value=(0.0, np.sum(yin)),
    )

    # calculate bin edges (of size len(xout)+1)
    xoutleft, xoutright = binedges(xout)
    xoutcdf = np.hstack([xoutleft, xoutright[-1]])

    # interpolate the CDF onto those bin edges
    cdfout = cdfinterpolator(xoutcdf)

    # take  derivative of the CDF to get flux per resampled bin
    # (xout is bin center, and yout is the flux in that bin)
    yout = np.diff(cdfout)

    if visualize:
        fi, (ax_cdf, ax_pdf) = plt.subplots(2, 1, sharex=True, figsize=(9, 6))
        inkw = dict(
            color="black",
            alpha=1,
            linewidth=3,
            marker=".",
            markeredgecolor="none",
        )
        outkw = dict(
            color="darkorange",
            alpha=1,
            linewidth=1,
            marker=".",
            markersize=8,
            markeredgecolor="none",
        )

        legkw = dict(
            fontsize=10,
            frameon=False,
            loc="upper left",
            bbox_to_anchor=(1, 1),
        )

        # plot the PDFs
        plt.sca(ax_pdf)
        plt.ylabel("Flux per (Original) Pixel")
        plt.xlabel("Pixel")
        # plot the original pixels (in df/dpixel to compare with resampled)
        plotboxy(xin, yin / xinbinsize, label="Original Pixels", **inkw)

        # what would a bad interpolation look like?
        badinterpolation = scipy.interpolate.interp1d(
            xin,
            yin / xinbinsize,
            kind="linear",
            bounds_error=False,
            fill_value=0.0,
        )
        plt.plot(
            xout,
            badinterpolation(xout),
            color="cornflowerblue",
            alpha=1,
            linewidth=1,
            marker=".",
            markersize=8,
            markeredgecolor="none",
            label="Silly Simple Interpolation",
        )

        # plot the flux-conserving resampled data (again, in df/d"pixel")
        plt.plot(
            xout, yout / xoutbinsize, label="Flux-Conserving Interpolation", **outkw
        )
        plt.legend(**legkw)

        # plot the CDFs
        plt.sca(ax_cdf)
        plt.ylabel("Cumulative Flux (from left)")

        # plot the original CDF
        plt.plot(xinforcdf, cdfin, label="Original Pixels", **inkw)

        # plot the interpolated CDF
        plt.plot(xoutcdf, cdfout, label="Flux-Conserved Resample", **outkw)
        if pause:
            a = input(
                "Pausing a moment to check on interpolation; press return to continue."
            )

        plt.tight_layout(rect=[0.0, 0.0, 0.67, 1])
        print("{:>6} = {:.5f}".format("Actual", np.sum(yin)))
        print(
            "{:>6} = {:.5f}".format(
                "Silly",
                np.sum(badinterpolation(xout) * xoutbinsize),
            )
        )
        print("{:>6} = {:.5f}".format("CDF", np.sum(yout)))

    # return the resampled y-values
    return yout


def bintogrid(
    x,
    y,
    unc=None,
    newx=None,
    dx=None,
    weighting="inversevariance",
    drop_nans=True,
):
    """
    Bin any x and y array onto a linearly uniform grid.

    Parameters
    ----------

    x : np.array
        The original independent variable.
        (For a spectrum example = wavelength)
    y : np.array
        The original dependent variable (same size as x).
        (For a spectrum example = flux)
    unc : np.array, or None
        The unceratinty on the dependent variable
        (For a spectrum example = the flux uncertainty)
    dx : np.array
        The fixed spacing for creating a new, linearly uniform
        grid that start at the first value of x. This will
        be ignored if `newx` != None.
    newx : np.array
        A new custom grid onto which we should bin.
    weighting : str
        How should we weight values when averaging
        them together into one larger bin?
        `weighting = 'inversevariance'`
            weights = 1/unc**2
         `weighting = {literally anything else}`
            uniform weights
        This will have no impact if `unc == None`, or for any
        new bins that effectively overlap less than one original
        unbinned point.
    drop_nans : bool
        Should we skip any bins turn out to be nans?
        This most often happens when bins are empty.

    Returns
    -------
    newx : np.array
        The new grid of x values.
    newy : np.array
        The new gridded y values.
    newunc : np.array
        The new gridded y uncertainties.
        These will be returned only if `unc != None` in the
        original inputs. Otherwise, only two outputs will be
        returned, newx and newy.

    # TODO: confirm nans in input arrays are handled OK
    # TODO: confirm units in y and unc are handled OK
    """

    try:
        x_unit = x.unit
        x_without_unit = x.value
    except AttributeError:
        x_unit = 1
        x_without_unit = x

    try:
        y_unit = y.unit
        y_without_unit = y.value
    except AttributeError:
        y_unit = 1
        y_without_unit = y

    # make up a grid, if one wasn't specified
    if newx is None:
        dx_without_unit = u.Quantity(dx).to(x_unit).value
        newx_without_unit = np.arange(
            np.nanmin(x_without_unit),
            np.nanmax(x_without_unit) + dx_without_unit,
            dx_without_unit,
        )
    else:
        newx_without_unit = u.Quantity(newx).to(x_unit).value

    # don't complain about zero-divisions in here (to allow infinite uncertainties)
    with np.errstate(divide="ignore", invalid="ignore"):

        # calculate weight integrals for the bin array
        ok = np.isnan(y_without_unit) == False

        # resample the sums onto that new grid
        if unc is None:
            weights = np.ones_like(x_without_unit)
        else:
            if weighting == "inversevariance":
                weights = 1 / unc ** 2
            else:
                weights = np.ones_like(x_without_unit)

            # ignore infinite weights (= 0 uncertainties)
            ok *= np.isfinite(weights)

        if np.any(ok):
            # TO-DO: check this nan handling on input arrays is OK?
            numerator = fluxconservingresample(
                x_without_unit[ok], (y_without_unit * weights)[ok], newx_without_unit
            )
            denominator = fluxconservingresample(
                x_without_unit[ok], (weights)[ok], newx_without_unit
            )

            # the binned weighted means on the new grid
            newy = numerator / denominator

            # the standard error on the means, for those bins
            newunc = np.sqrt(1 / denominator)
        else:
            newy = np.nan * newx_without_unit
            newunc = np.nan * newx_without_unit

    # remove any empty bins
    if drop_nans:
        ok = np.isfinite(newy)
    else:
        ok = np.ones_like(newx_without_unit).astype(bool)

    # if no uncertainties were given, don't return uncertainties
    if unc is None:
        return newx_without_unit[ok] * x_unit, newy[ok] * y_unit
    else:
        return newx_without_unit[ok] * x_unit, newy[ok] * y_unit, newunc[ok] * y_unit
        # TODO: check units on uncertainties


def bintoR(
    x, y, unc=None, R=50, xlim=None, weighting="inversevariance", drop_nans=True
):
    """
    Bin any x and y array onto a logarithmicly uniform grid,
    characterized by

    Parameters
    ----------

    x : np.array
        The original independent variable.
        (For a spectrum example = wavelength)
    y : np.array
        The original dependent variable (same size as x).
        (For a spectrum example = flux)
    unc : np.array, or None
        The unceratinty on the dependent variable
        (For a spectrum example = the flux uncertainty)
    R : np.array
        The spectral resolution R=x/dx for creating a new,
        logarithmically uniform grid that starts at the first
        value of x.
    xlim : list, np.array
        A two-element list indicating the min and max values of
        x for the new logarithmically spaced grid. If None,
        these limits will be created from the data themselves
    weighting : str
        How should we weight values when averaging
        them together into one larger bin?
        `weighting = 'inversevariance'`
            weights = 1/unc**2
         `weighting = {literally anything else}`
            uniform weights
        This will have no impact if `unc == None`, or for any
        new bins that effectively overlap less than one original
        unbinned point.
    drop_nans : bool
        Should we skip any bins turn out to be nans?
        This most often happens when bins are empty.

    Returns
    -------
    newx : np.array
        The new grid of x values.
    newy : np.array
        The new gridded y values.
    newunc : np.array
        The new gridded y uncertainties.
        These will be returned only if `unc != None` in the
        original inputs. Otherwise, only two outputs will be
        returned, newx and newy.

    # TODO: confirm nans in input arrays are handled OK
    """

    try:
        x_unit = x.unit
        x_without_unit = x.value
    except AttributeError:
        x_unit = 1
        x_without_unit = x

    # create a new grid of x at the given resolution
    lnx = np.log(x_without_unit)
    dnewlnx = 1.0 / R

    # set the limits of the new xgrid (in log space)
    if xlim is None:
        # use the input grid to set the limits
        lnxbottom, lnxtop = np.nanmin(lnx), np.nanmax(lnx)
    else:
        # use the custom xlim to set the limits
        lnxbottom, lnxtop = xlim

    # create a new, log-uniform grid of x values
    newlnx = np.arange(lnxbottom, lnxtop + dnewlnx, dnewlnx)

    # now do the binning on a uniform grid of lnx
    if unc is None:
        blnx, by = bintogrid(
            lnx, y, unc, newx=newlnx, weighting=weighting, drop_nans=drop_nans
        )
        return np.exp(blnx) * x_unit, by
    else:
        blnx, by, bunc = bintogrid(
            lnx, y, unc, newx=newlnx, weighting=weighting, drop_nans=drop_nans
        )
        return np.exp(blnx) * x_unit, by, bunc



from ..rainbows import *
from .setup_tests import *


def test_rainbow_npy():
    filename = os.path.join(test_directory, "test.rainbow.npy")
    a = SimulatedRainbow()
    a.save(filename)
    b = Rainbow(filename)
    assert a == b


def test_rainbow_FITS():
    filename = os.path.join(test_directory, "test.rainbow.fits")
    a = SimulatedRainbow()
    a.save(filename, overwrite=True)
    b = Rainbow(filename)
    assert a == b


def test_text():
    filename = os.path.join(test_directory, "test.rainbow.txt")
    a = SimulatedRainbow()
    a.save(filename)
    b = Rainbow(filename)
    # assert a == b


def test_guess_readers():

    assert guess_reader("some-neat-file.rainbow.npy") == from_rainbow_npy
    assert guess_reader("some-neat-file-x1dints.fits") == from_x1dints
    assert guess_reader("S3_neat-planet_Save.txt") == from_eureka
    assert guess_reader("some-weird-file.txt") == from_text


def test_guess_writers():

    assert guess_writer("some-neat-file.rainbow.npy") == to_rainbow_npy
